// **************************************************************************************
// ** 
// **  RGBW DCC LocoNet + Arduino
// **  Versie 1, Huib Maaskant
// ** 
// **  RGB LEDs:
// **  Pin 3 --> R
// **  Pin 5 --> G
// **  Pin 6 --> B
// **  Pin 11 --> W
// **
// **  LocoNet:
// **  Pin 8 --> LocoNet Rx
// **  Pin 12 --> LocoNet Tx
// ** 
// **  DCC:
// **  Pin 2 --> DCC Rx
// **
// **  Overig:
// **  Pin 13 --> BUSY LED
// **  Vrije digitale pinnen: 0, 1, 4, 7, 9, 10
// **  Let op: door de LocoNet library is geen PWM meer beschikbaar op pin 9 en 10
// **  
// **  Adressen (vastgelegd in BASISADRES; gelijk voor DCC en LocoNet):
// **  BASISADRES    R      Preset 1 (dag)
// **  BASISADRES    G      Preset 2 (avond)
// **  BASISADRES+1  R      Preset 3 (nacht)
// **  BASISADRES+1  G      Preset 4 (ochtend)
// **  BASISADRES+2  R      Alles uit
// **  BASISADRES+2  G      Vol wit aan
// ** 
// **  BASISADRES+3  R / G  Storing (geel) uit / aan
// **  BASISADRES+4  R / G  Storing (rood) uit / aan
// **  BASISADRES+5  R / G  Karaoke/disco uit / aan
// **  BASISADRES+6  R / G  Onweer uit / aan
// ** 
// **  BASISADRES+7  R / G  Rood - / +
// **  BASISADRES+8  R / G  Groen - / +
// **  BASISADRES+9  R / G  Blauw - / +
// **  BASISADRES+10 R / G  Wit - / +
// **************************************************************************************

// Loconet library
#include <LocoNet.h>
#define LOCONET_TX_PIN 12
lnMsg *LnPacket;

// DCC library
#include "NmraDcc.h"
#define DCC_RX_PIN 2
NmraDcc  Dcc;
DCC_MSG  Packet;

// RGBW pins
#define RED 3
#define GREEN 5 
#define BLUE 6 
#define WHITE 11

// BUSY LED
#define STATUSLED 13

// Adres
#define BASISADRES 101        // <-- Pas aan om de adressen te veranderen

// DIM factor voor felle LEDstrips, waarde 0.1 - 1.0, waarbij 1.0 niet gedimd is
#define DIMFACTOR 0.8

// Presets dag / avond / nacht / ochtend
#define PRESET1_R DIMFACTOR*76         // Dag
#define PRESET1_G DIMFACTOR*93   
#define PRESET1_B DIMFACTOR*79
#define PRESET1_W DIMFACTOR*214

#define PRESET2_R DIMFACTOR*255        // Avond
#define PRESET2_G DIMFACTOR*56
#define PRESET2_B DIMFACTOR*33
#define PRESET2_W DIMFACTOR*52

#define PRESET3_R DIMFACTOR*0          // Nacht  
#define PRESET3_G DIMFACTOR*0
#define PRESET3_B DIMFACTOR*40
#define PRESET3_W DIMFACTOR*3

#define PRESET4_R DIMFACTOR*243        // Ochtend
#define PRESET4_G DIMFACTOR*62
#define PRESET4_B DIMFACTOR*4
#define PRESET4_W DIMFACTOR*20

// Preset type animatie
#define ANIMNONE 0           // Geen animatie
#define ANIMKARA 1           // Karaoke
#define ANIMFLASH1 2         // Onweer fase 1: donker worden
#define ANIMFLASH2 3         // Onweer fase 2: bliksem

#define TRANSSTEPS 100       // Aantal stappen in de transitie van A naar B
#define TRANSLOOPSKIP 50     // Aantal malen dat per stap de transitieloop in loop() wordt overgeslagen >> pas deze waarde aan om de transitie sneller of langzamer te maken
#define ANIMKARALOOPSKIP 5   // Idem voor de Karaoke animatie  >> pas deze waarde aan om de animatie sneller of langzamer te maken
#define ANIMFLASHLOOPSKIP 5  // Idem voor de onweer animatie >> pas deze waarde aan om de animatie sneller of langzamer te maken

// Variabelen voor LEDS
int red = 0;                // De actuele kleurwaarden op de RGBW uitgangen 
int green = 0;
int blue = 0;
int white = 0;

int a_red = 0;              // Transitie van kleur A ...
int a_green = 0;
int a_blue = 0;
int a_white = 0;

int b_red = 0;              // ... naar kleur B.
int b_green = 0;
int b_blue = 0;
int b_white = 0;

int s_red = 0;              // Tijdens effecten(animaties) bewaren we hierin de laatste stand zodat naar de oude waarden terug kunnen
int s_green = 0;
int s_blue = 0;
int s_white = 0;

// Overige globale variabelen
bool transrunning = false;  // Is er een transitie bezig?
bool animrunning = false;   // Is er een animatie bezig?
int animtype = ANIMNONE;    // Als er een animatie loopt, welke is dat dan?
int stepcount = 0;          // Teller die de stappen van de transitie of animatie telt
int loopskip = 0;           // Teller voor het skippen van de transitie of animatie routine in loop()
int dccdeaf = 0;            // Teller om tijdelijk 'doof' te zijn voor dubbele DCC commando's


// ********************************************************************************
// ** 
// **  Hoofdroutines
// ** 
// ********************************************************************************

// Opstarten
void setup() {
  // LocoNet interface starten
  LocoNet.init(LOCONET_TX_PIN);

  // DCC interface starten
  // Setup which External Interrupt
  Dcc.pin(0, DCC_RX_PIN, 1);
  
  // Call the main DCC Init function to enable the DCC Receiver
  Dcc.init( MAN_ID_DIY, 10, CV29_ACCESSORY_DECODER | CV29_OUTPUT_ADDRESS_MODE, 0 );

  // Seriele poort openen
  Serial.begin(115200);
  Serial.println(" ");
  Serial.println("---------------------------");
  Serial.println("RGBW DCC LocoNet decoder...");
  Serial.println("---------------------------");
  Serial.print("Basisadres: ");
  Serial.println(BASISADRES);
  Serial.println("---------------------------");
  Serial.println(" ");

  // LEDS instellen
  pinMode(RED, OUTPUT);
  pinMode(GREEN, OUTPUT);
  pinMode(BLUE, OUTPUT);
  pinMode(WHITE, OUTPUT);

  SetRGBW();

  pinMode(STATUSLED, OUTPUT);
  digitalWrite(STATUSLED, LOW);

  // Licht infaden naar de dagstand
  StartTransition(PRESET1_R, PRESET1_G, PRESET1_B, PRESET1_W);
}

// Hoofdlus
void loop() {
  // Check voor ontvangen LocoNet pakketjes
  LnPacket = LocoNet.receive();
  
  // Als er een pakketje is, worden automatisch de functies verderop aangeroepen door LocoNet.processSwitchSensorMessage
  if (LnPacket) {
    LocoNet.processSwitchSensorMessage(LnPacket);
  }
  // Als het pakket een commnando voor ons bevatte, is dat inmiddels verwerkt

  // Verwerk het DCC signaal
  Dcc.process();
  // Als er een commando binnen is, wordt het nu verwerkt door notifyDccAccTurnoutOutput


  // Transities een stap verder brengen
  if (transrunning) {
    if (loopskip <1) {
      red = a_red+((b_red - a_red)*stepcount/TRANSSTEPS);
      green = a_green+((b_green - a_green)*stepcount/TRANSSTEPS);
      blue = a_blue+((b_blue - a_blue)*stepcount/TRANSSTEPS);
      white = a_white+((b_white - a_white)*stepcount/TRANSSTEPS);
      SetRGBW();

      stepcount ++;
      if (stepcount > TRANSSTEPS) { // Dan zijn we klaar
        transrunning = false;
        digitalWrite(STATUSLED, LOW);
        s_red = red;
        s_green  = green;
        s_blue = blue;
        s_white = white;
      }
      loopskip = TRANSLOOPSKIP;
    }
    else {
      loopskip --;
    }
  }

  // Animaties een stap verder brengen
  if (animrunning) {
    if (animtype == ANIMKARA) { // De karaoke animatie
      if (loopskip < 1) {
        if (stepcount == 0) {   // Bedenk een kleur en start een overgang naar die kleur
          a_red = red;
          a_green = green;
          a_blue = blue;
          a_white = white;

          b_red = random(2) * 200;
          b_green = random(2) * 200;
          b_blue = random(2) * 200;
          b_white = 0;

          stepcount = 1;
        }
        red = a_red+((b_red - a_red)*stepcount/TRANSSTEPS);
        green = a_green+((b_green - a_green)*stepcount/TRANSSTEPS);
        blue = a_blue+((b_blue - a_blue)*stepcount/TRANSSTEPS);
        white = a_white+((b_white - a_white)*stepcount/TRANSSTEPS);
        SetRGBW();

        stepcount ++;
        if (stepcount > TRANSSTEPS) { // Dan zijn we klaar met deze transitie en moeten we een nieuwe kleur kiezen
          stepcount = 0;
        }
        loopskip = ANIMKARALOOPSKIP;  
      }
      else {
          loopskip --;
      }

    }
    
    if (animtype == ANIMFLASH1) { // De onweeranimatie fase 1: donker worden
      if (loopskip <1) {
        red = a_red+((b_red - a_red)*stepcount/TRANSSTEPS);
        green = a_green+((b_green - a_green)*stepcount/TRANSSTEPS);
        blue = a_blue+((b_blue - a_blue)*stepcount/TRANSSTEPS);
        white = a_white+((b_white - a_white)*stepcount/TRANSSTEPS);
        SetRGBW();

        stepcount ++;
        if (stepcount > TRANSSTEPS) { // Dan zijn we klaar en kunnen we door naar fase 2
          animtype = ANIMFLASH2;
          stepcount = 0;
        }
        loopskip = TRANSLOOPSKIP;
      }
      else {
        loopskip --;
      }
    }

    if (animtype == ANIMFLASH2) { // De onweeranimatie fase 2: bliksemen
      if (loopskip < 1) {      
        if (stepcount < 1) {    // Dan is het tijd om wat te veranderen
          if (white < 255 ) {    // Dan gaat nu de bliksem aan
            white = 255;
            blue = 255;
            SetRGBW();
            stepcount = 2+random(5)*2;  // Lengte van de flits
          }
          else {
            white = b_white;
            blue = b_blue;
            SetRGBW();
            stepcount = random(15)*80; // Lengte van de pauze tussen de flitsen
          }
        }
        else {
          stepcount--;
        }
        loopskip = ANIMFLASHLOOPSKIP;
      }
      else {
          loopskip --;
      }
    }
  }
  
  if (dccdeaf > 0) {dccdeaf--;}
  delay (1);
}


// ********************************************************************************
// ** 
// **  Functies
// ** 
// ********************************************************************************

// Routine voor waarde zichtbaar maken in LEDS + melding via serial monitor
// Gebruikt de globale variabelen red, green, blue, white
void SetRGBW () {

  // Check op waarde. Wellicht overbodig, maar better safe than sorry
  if (red < 0) {red = 0;}
  if (red > 255) {red = 255;}
  if (green < 0) {green = 0;}
  if (green > 255) {green = 255;}
  if (blue < 0) {blue = 0;}
  if (blue > 255) {blue = 255;}
  if (white < 0) {white = 0;}
  if (white > 255) {white = 255;}

  // Waarde op de PWM pinnen zetten
  analogWrite(RED, red);
  analogWrite(GREEN, green);
  analogWrite(BLUE, blue);
  analogWrite(WHITE, white);

  Serial.print(red);
  Serial.print(":");
  Serial.print(green);
  Serial.print(":");
  Serial.print(blue);
  Serial.print(":");
  Serial.println(white);
}

// Routine om een transitie te starten
// Transitie is van de huidige waarden in de globale variabelen red, green, blue, white
// Naar de vier argumenten van deze functie
void StartTransition (int tgt_red, int tgt_green, int tgt_blue, int tgt_white) {

  digitalWrite(STATUSLED, HIGH);
  
  a_red = red;        // Startpunt (A) is de huidige waarde van de LEDs
  a_green = green;
  a_blue = blue;
  a_white = white;

  b_red = tgt_red;   // Eindpunt (B) is de argumenten van deze functie
  b_green = tgt_green;
  b_blue = tgt_blue;
  b_white = tgt_white;

  stepcount = 1;
  loopskip = 0;
  transrunning = true;
  // Vanaf nu gaat de loop() functie de transitie maken
}

// Stopt alle lopende transities en animaties
void StopAllTransAnim () {
  transrunning = false;
  animrunning = false;
  digitalWrite(STATUSLED, LOW);
}

// Routine voor verwerken commando's die via LocoNet of DCC worden ontvangen
// address: het adres natuurlijk
// direction: ROOD == 0, GROEN != 0
void ProcessCommand (uint16_t address, uint8_t direction) {

  switch(address){
    case BASISADRES:
      if(direction == 0){     // R: Preset 1
        StopAllTransAnim ();
        StartTransition(PRESET1_R, PRESET1_G, PRESET1_B, PRESET1_W);
        Serial.println("Preset 1");
      } else {                // G: Preset 2
        StopAllTransAnim ();
        StartTransition(PRESET2_R, PRESET2_G, PRESET2_B, PRESET2_W);
        Serial.println("Preset 2");
      }
      break;

    case BASISADRES+1:
      if(direction ==0){     // R: Preset 3
        StopAllTransAnim ();
        StartTransition(PRESET3_R, PRESET3_G, PRESET3_B, PRESET3_W);
        Serial.println("Preset 3");
      } else {               // G: Preset 4
        StopAllTransAnim ();
        StartTransition(PRESET4_R, PRESET4_G, PRESET4_B, PRESET4_W);
        Serial.println("Preset 4");
      }
      break;

    case BASISADRES+2:
      if(direction ==0){     // R: Alles uit
        StopAllTransAnim ();
        StartTransition(0, 0, 0, 0);
        Serial.println("Alles uit");
      } else {               // G: Alleen wit vol aan
        StopAllTransAnim ();
        StartTransition(0, 0, 0, 255);
        Serial.println("Wit vol aan");
      }
      break;

    case BASISADRES+3:
      if(direction ==0){     // R: Storingslicht geel uit (terug naar normaal)
        StopAllTransAnim ();
        StartTransition (s_red, s_green, s_blue, s_white);
        Serial.println("Storingslicht geel uit");
      } else {               // G: Storingslicht geel aan
        StopAllTransAnim ();
        s_red = red;
        s_green = green;
        s_blue = blue;
        s_white = white;
        red = 255;
        green = 255;
        blue = 0;
        white = 0;
        SetRGBW();
        Serial.println("Storingslicht geel aan");
      }
      break;

    case BASISADRES+4:
      if(direction ==0){     // R: Storingslicht rood uit (terug naar normaal)
        StopAllTransAnim ();
        StartTransition (s_red, s_green, s_blue, s_white);
        Serial.println("Storingslicht rood uit");
      } else {               // G: Storingslicht rood aan
        StopAllTransAnim ();
        s_red = red;
        s_green = green;
        s_blue = blue;
        s_white = white;
        red = 255;
        green = 0;
        blue = 0;
        white = 0;
        SetRGBW();
        Serial.println("Storingslicht rood aan");
      }
      break;

    case BASISADRES+5:
      if(direction ==0){     // R: Karaoke uit
        if (animrunning) {
          StopAllTransAnim ();
          StartTransition (s_red, s_green, s_blue, s_white);
          Serial.println("Karaoke uit");
        }
      } else {               // G: Karaoke aan
        StopAllTransAnim ();
        s_red = red;
        s_green = green;
        s_blue = blue;
        s_white = white;
        
        digitalWrite(STATUSLED, HIGH);
        animtype = ANIMKARA;
        stepcount = 0;
        loopskip = 0;
        animrunning = true;
        
        Serial.println("Karaoke aan");
      }
      break;

    case BASISADRES+6:
      if(direction ==0){     // R: Onweer uit
        if (animrunning) {
          StopAllTransAnim ();
          StartTransition (s_red, s_green, s_blue, s_white);
          Serial.println("Onweer uit");
        }
      } else {               // G: Onweer aan
        StopAllTransAnim ();
        s_red = red;
        s_green = green;
        s_blue = blue;
        s_white = white;

        a_red = red;        // Startpunt (A) is de huidige waarde van de LEDs
        a_green = green;
        a_blue = blue;
        a_white = white;

        b_red = 0.4*red;   // Eindpunt (B) is een donkere lucht
        b_green = 0.4*green;
        b_blue = 0.4*blue;
        b_white = 0.4*white;

        stepcount = 1;
        loopskip = 0;
        animrunning = true;
        animtype = ANIMFLASH1;        
        digitalWrite(STATUSLED, HIGH);

        Serial.println("Onweer aan");
      }
      break;

    case BASISADRES+7:
      if(direction ==0){     // R: Rood -
        StopAllTransAnim ();
        if (red >= 5) {red = red - 5;}
        SetRGBW();
        Serial.println("Rood -");
      } else {               // G: Rood +
        StopAllTransAnim ();
        if (red <= 250) {red = red + 5;}
        SetRGBW();
        Serial.println("Rood +");
      }
      break;

    case BASISADRES+8:
      if(direction ==0){     // R: Groen -
        StopAllTransAnim ();
        if (green >= 5) {green = green - 5;}
        SetRGBW();
        Serial.println("Groen -");
      } else {               // G: Groen +
        StopAllTransAnim ();
        if (green <= 250) {green = green + 5;}
        SetRGBW();
        Serial.println("Groen +");
      }
      break;

    case BASISADRES+9:
      if(direction ==0){     // R: Blauw -
        StopAllTransAnim ();
        if (blue >= 5) {blue = blue - 5;}
        SetRGBW();
        Serial.println("Blauw -");
      } else {               // G: Blauw +
        StopAllTransAnim ();
        if (blue <= 250) {blue = blue + 5;}
        SetRGBW();
        Serial.println("Blauw +");
      }
      break;

    case BASISADRES+10:
      if(direction ==0){     // R: Wit -
        StopAllTransAnim ();
        if (white >= 5) {white = white - 5;}
        SetRGBW();
        Serial.println("Wit -");
      } else {               // G: Wit +
        StopAllTransAnim ();
        if (white <= 250) {white = white + 5;}
        SetRGBW();
        Serial.println("Wit +");
      }
      break;

    default:
      break;
    }

}


// ********************************************************************************
// ** 
// **  DCC
// ** 
// ********************************************************************************

// This function is called whenever a normal DCC Turnout Packet is received and we're in Output Addressing Mode
// Hierin handelen we dus de binnenkomende commando's af
// DCC verstuurt hetzelfde commando diverse keren kort achter elkaar
// Daarom moeten we iets inbouwen om na het ontvangen van een passend commando tijdelijk 'doof' te worden
void notifyDccAccTurnoutOutput( uint16_t Addr, uint8_t Direction, uint8_t OutputPower )
{
  if (OutputPower !=0 && Addr >= BASISADRES && Addr <= BASISADRES+10) {
      if (dccdeaf == 0) {
        ProcessCommand (Addr, Direction);
        dccdeaf = 500;
      }
  }
}


// ********************************************************************************
// ** 
// **  LocoNet
// ** 
// ********************************************************************************

// This call-back function is called from LocoNet.processSwitchSensorMessage
// for all Switch Request messages
// Address: het adres natuurlijk
// Direction: ROOD == 0, GROEN != 0
// Output: Aan !=0, Uit == 0 - we kijken alleen naar de Aan stand
void notifySwitchRequest( uint16_t Address, uint8_t Output, uint8_t Direction ) {
  
  if (Output !=0) {
    ProcessCommand (Address, Direction);
  }
}


// This call-back function is called from LocoNet.processSwitchSensorMessage
// for all Power messages
// Oftewel: staat de hele handel op stop?
void notifyPower(uint8_t State) {
  // State GO: !=0, STOP: == 0

  if (State) {
      Serial.println("[LocoNet GO]");
      StopAllTransAnim();
      StartTransition (s_red, s_green, s_blue, s_white);
  }
  else {
      Serial.println("[LocoNet STOP]");
      StopAllTransAnim();
      s_red = red;
      s_green = green;
      s_blue = blue;
      s_white = white;
      red = 255;
      green = 0;
      blue = 0;
      white = 0;
      SetRGBW();
  }
}

// Overige LocoNet functies, niet gebruikt

// This call-back function is called from LocoNet.processSwitchSensorMessage
// for all Switch Output Report messages
void notifySwitchOutputsReport( uint16_t Address, uint8_t ClosedOutput, uint8_t ThrownOutput) {
}

// This call-back function is called from LocoNet.processSwitchSensorMessage
// for all Switch Sensor Report messages
void notifySwitchReport( uint16_t Address, uint8_t State, uint8_t Sensor ) {
}

// This call-back function is called from LocoNet.processSwitchSensorMessage
// for all Switch State messages
void notifySwitchState( uint16_t Address, uint8_t Output, uint8_t Direction ) {
}

// This call-back function is called from LocoNet.processSwitchSensorMessage
// for all MultiSensePower messages
void notifyMultiSensePower(uint8_t BoardID, uint8_t Subdistrict, uint8_t Mode, uint8_t Direction) {
}

// This call-back function is called from LocoNet.processSwitchSensorMessage
// for all notifyMultiSenseTransponder messages
void notifyMultiSenseTransponder(uint16_t Address, uint8_t Zone, uint16_t LocoAddress, uint8_t Present) {
}

// This call-back function is called from LocoNet.processSwitchSensorMessage
// for all LongAck messages
void notifyLongAck(uint8_t d1, uint8_t d2) {
}
